package com.esdtwo.objectdetection.lab3;

import java.io.File;
import java.lang.invoke.MethodHandles;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;

import org.bytedeco.opencv.opencv_core.*;
import static org.bytedeco.opencv.global.opencv_core.*;
import static org.bytedeco.opencv.global.opencv_imgproc.*;
import static org.bytedeco.opencv.global.opencv_imgcodecs.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class Cli
{
	private static final String version = "1.0.0";

	//Print to user while also printing to file
	private static final Logger logger = LoggerFactory.getLogger(MethodHandles.lookup().lookupClass());

	private static DateTimeFormatter fileDateTime = DateTimeFormatter.ofPattern("yyyy-MM-dd_HH.mm.ss");

	//Experimentally determined values for the min and max for thresholding
	private static final double lowHue = 0;
	private static final double lowSat = 0;
	private static final double lowVal = 97;
	private static final Mat minValues = new Mat(lowHue, lowSat, lowVal,0);

	private static final double highHue = 117.0 / 2.0;
	private static final double highSat = 136;
	private static final double highVal = 255;
	private static final Mat maxValues = new Mat(highHue, highSat, highVal,0);

	/** Distance between the cameras in millimeters; provided*/
	private static final double baseline = 60.0; 
	/** Focal length of the cameras; provided*/
	private static final double focalLength = 6.0;
	/**Pixel size, measured in millimeters per pixel*/
	private static final double pixelSize = 0.006;

	//Centre values for a given image. Images are a fixed resolution of 752x480
	private static final double xCentre = 752.0 / 2.0;
	private static final double yCentre = 480.0 / 2.0;

	/**Storage for centroids of objects from the left camera*/
	private static List<Point> leftCentroids = new ArrayList<>();
	/**Storage for centroids of objects from the right camera*/
	private static List<Point> rightCentroids = new ArrayList<>();

	/**
	 * Static blocks are only executed once, and are always executed first.
	 * Always print version information.
	 */
	static
	{
		logger.info("================");
		logger.info("Lab 3");
		logger.info("Version: " + version);
		logger.info("================");
	}

	public static void main(String[] args)
	{
		//Massive try/catch block, to catch any extraneous error without crashing
		try{
			//For each file in the current directory...
			for(File file : new File(".").listFiles()){
				//Ignore all non-JPG files
				if (!file.getName().contains("jpg")) { continue; }
				//Print filename and path
				logger.info("New file");
				logger.debug("File detected as: " + file.getAbsolutePath());

				//Read in file as BGR image
				Mat readImage = imread(file.getAbsolutePath());

				//Dump basic stats about the file to log
				logger.trace("File: " + readImage.toString());

				if(!readImage.empty()){
					//Output object (empty, for now)
					Mat blurredImage = new Mat();
					logger.info("Blurring image...");
					blur(readImage,blurredImage,new Size(7,7));
					Mat colourConversion = new Mat();
					logger.info("Converting to HSV...");
					cvtColor(blurredImage,colourConversion,COLOR_BGR2HSV);
					Mat thresholdedImage = new Mat();

					logger.info("Creating edge mask...");
					//Advanced threshold; white if pixel is within values, black if not
					inRange(colourConversion, minValues, maxValues, thresholdedImage);

					//Erosion and Dilation tutorial: 
					//https://docs.opencv.org/3.4/db/df6/tutorial_erosion_dilatation.html
					//TL;DR: de-noising, accentuates traits we're already looking for
					Mat ellipse = getStructuringElement(MORPH_ELLIPSE, new Size(12,12));
					Mat rectangle = getStructuringElement(MORPH_RECT, new Size(12,12));
					Mat erodedImage = new Mat();
					Mat dilatedImage = new Mat();
					logger.info("Eroding and dilating to reduce noise...");
					morphologyEx(thresholdedImage,erodedImage,MORPH_CLOSE,rectangle);
					erode(erodedImage,erodedImage,rectangle);
					erode(erodedImage,erodedImage,rectangle);
					erode(erodedImage,erodedImage,ellipse);
					erode(erodedImage,erodedImage,ellipse);
					dilate(erodedImage,dilatedImage,ellipse);
					dilate(dilatedImage,dilatedImage,ellipse);
					dilate(dilatedImage,dilatedImage,ellipse);
					dilate(dilatedImage,dilatedImage,ellipse);
					dilate(dilatedImage,dilatedImage,ellipse);

					//"Contours": edges of visible objects
					MatVector contours = new MatVector();
					Mat heirarchy = new Mat();
					logger.info("Finding edges...");
					findContours(dilatedImage, contours, heirarchy, RETR_TREE, CHAIN_APPROX_SIMPLE);
					//Dump contour and heirarchy information to file
					logger.trace("Contours: " + contours.toString());
					logger.trace("heirarchy: " + heirarchy.toString());
					logger.info("drawing edges...");
					//Draw all contours on top of the original image, in blue
					drawContours(readImage,contours,-1,new Scalar(255,0,0,0));

					int i = 0;
					//If we don't have any contours, back out now
					if (contours.size() > 0){
						logger.info("Drawing centroids...");
						//For all the contours, assuming they exist...
						for(Mat contour : contours.get()){
							if(!contour.empty()){
								//Dump contour to log
								logger.trace(contour.toString());
								Mat approxContour = new Mat();
								//Shrink number of vertices in the contour
								approxPolyDP(contour,approxContour,(0.01*arcLength(contour,true)),true);
								//Get area of the contour
								double area = contourArea(contour);
								logger.trace(approxContour.toString());
								logger.info("Contour area: " + area);
								logger.info("Contour segments: " + approxContour.arrayHeight());
								//Only use contours with a large amount of 
								if (approxContour.arrayHeight() > 8 && area > 8000){
									logger.warn("Found valid contour!");
									i++;
									Moments centroid = moments(contour);
									double cX = centroid.m10() / centroid.m00();
									double cY = centroid.m01() / centroid.m00();
									Point centre = new Point((int)cX,(int)cY);
									logger.info("Centroid " + i + ": (" + cX + "," + cY + ")");
									if(file.getName().contains("left")){
										leftCentroids.add(centre);
									} else if(file.getName().contains("right")){
										rightCentroids.add(centre);
									}

									circle(readImage,centre,2,new Scalar(0,0,255,0));
								}
							}
						}
					} else {
						logger.error("No contours found!");
					}

					logger.info("Saving to file...");
					i = 0;
					for(Mat contour : contours.get()){
						logger.trace(contour.toString());
						if (!contour.empty()){
							//saveImage(contour,file.getParent(),"final-"+i);
						}
						i++;
					}
					saveImage(readImage,file.getName(),"import-with-contours");
				} else{
					logger.error("Invalid file location!");
				}
				logger.info("-------------------------------------");
			}
			logger.warn("Coordinate format: (x,y,z)");
			for(int i = 0; i < leftCentroids.size(); i++){
				Point leftCentroid = leftCentroids.get(i);
				Point rightCentroid = rightCentroids.get(i);
				double xLeftOffset = xCentre - leftCentroid.x();
				double yLeftOffset = yCentre - leftCentroid.y();
				double xRightOffset = xCentre - rightCentroid.x();
				double yRightOffset = yCentre - rightCentroid.y();

				double disparity = (Math.abs(xLeftOffset - xRightOffset) * pixelSize);
				double z = (baseline * focalLength) / disparity;
				double finalX = (z * xLeftOffset * pixelSize) / focalLength;
				double finalY = (z * yLeftOffset * pixelSize) / focalLength;

				logger.info("Ball position in image number " + (i+5) + ": (" + finalX + ", " + finalY + ", " + z + ").");
			}
		}
		//Catch portion of the main try/catch block. Catch the error, print the error, gracefully close
		catch(Exception e){
			logger.error("ERROR CAUGHT - CLOSING PROGRAM.");
			logger.error("Error caught: ",e);
		}
	}

	/**
	 * Dumps image to file for viewing
	 *
	 * @param image		Image to be stored
	 * @param location	location to store the image in
	 * @param name		label to be appended to the filename
	 */
	private static void saveImage(Mat image, String location,String name){
		//Trim out the jpg portion. We don't need that information, and it will swell our dataset.
		String trimmedLocation = location.replaceAll(".jpg","");
		//Make sure to save to the current directory, and save as PNG
		String savedFile = trimmedLocation  + name + fileDateTime.format(LocalDateTime.now()) + ".png";
		//Dump image to file
		imwrite(savedFile,image);
	}
}
